import warnings
import numpy as np
from functools import reduce
from itertools import groupby
from aimsgb.utils import reduce_vector
from pymatgen.core.structure import Structure, Lattice, PeriodicSite
from pymatgen.transformations.advanced_transformations import CubicSupercellTransformation
from pymatgen.analysis.structure_matcher import StructureMatcher

__author__ = "Jianli CHENG and Kesong YANG"
__copyright__ = "Copyright 2018 University of California San Diego"
__maintainer__ = "Jianli CHENG"
__email__ = "jic198@ucsd.edu"
__status__ = "Production"
__date__ = "January 26, 2018"


class Grain(Structure):
    """
    We use the Structure class from pymatgen and add several new functions.
    """
    def __init__(self, lattice, species, coords, charge=None,
                 validate_proximity=False, to_unit_cell=False,
                 coords_are_cartesian=False, site_properties=None):

        super(Structure, self).__init__(lattice, species, coords, charge=charge,
                                        validate_proximity=validate_proximity,
                                        to_unit_cell=to_unit_cell,
                                        coords_are_cartesian=coords_are_cartesian,
                                        site_properties=site_properties)

        self._sites = list(self._sites)

    @staticmethod
    def get_b_from_a(grain_a, csl):
        grain_b = grain_a.copy()
        csl_t = csl.transpose()
        if sum(abs(csl_t[0]) - abs(csl_t[1])) > 0:
            axis = (1, 0, 0)
        else:
            axis = (0, 1, 0)
        anchor = grain_b.lattice.get_cartesian_coords(np.array([.0, .0, .0]))
        # print(axis, anchor)
        # exit()
        grain_b.rotate_sites(theta=np.radians(180), axis=axis, anchor=anchor)
        return grain_b

    def make_supercell(self, scaling_matrix):
        """
        Create a supercell. Very similar to pymatgen's Structure.make_supercell
        However, we need to make sure that all fractional coordinates that equal
        to 1 will become 0 and the lattice are redefined so that x_c = [0, 0, c]
    
        Args:
            scaling_matrix (3x3 matrix): The scaling matrix to make supercell.
        """
        s = self * scaling_matrix
        for i, site in enumerate(s):
            f_coords = np.mod(site.frac_coords, 1)
            # The following for loop is probably not necessary. But I will leave
            # it here for now.
            for j, v in enumerate(f_coords):
                if abs(v - 1) < 1e-6:
                    f_coords[j] = 0
            s[i] = PeriodicSite(site.specie, f_coords, site.lattice,
                                properties=site.properties)
        self._sites = s.sites
        self._lattice = s.lattice
        new_lat = Lattice.from_parameters(*s.lattice.parameters)
        self.lattice = new_lat

    def delete_bt_layer(self, bt, tol=0.25, axis=2):
        """
        Delete bottom or top layer of the structure.
        Args:
            bt (str): Specify whether it's a top or bottom layer delete. "b"
                means bottom layer and "t" means top layer.
            tol (float), Angstrom: Tolerance factor to determine whether two
                atoms are at the same plane.
                Default to 0.25
            axis (int): The direction of top and bottom layers. 0: x, 1: y, 2: z

        """
        if bt == "t":
            l1, l2 = (-1, -2)
        else:
            l1, l2 = (0, 1)

        l = self.lattice.abc[axis]
        layers = self.sort_sites_in_layers(tol=tol, axis=axis)
        l_dist = abs(layers[l1][0].coords[axis] - layers[l2][0].coords[axis])
        l_vector = [1, 1]
        l_vector.insert(axis, (l - l_dist) / l)
        new_lat = Lattice(self.lattice.matrix * np.array(l_vector)[:, None])

        layers.pop(l1)
        sites = reduce(lambda x, y: np.concatenate((x, y), axis=0), layers)
        new_sites = []
        l_dist = 0 if bt == "t" else l_dist
        l_vector = [0, 0]
        l_vector.insert(axis, l_dist)
        for i in sites:
            new_sites.append(PeriodicSite(i.specie, i.coords - l_vector,
                                          new_lat, coords_are_cartesian=True))
        self._sites = new_sites
        self._lattice = new_lat

    def sort_sites_in_layers(self, tol=0.25, axis=2):
        """
        Sort the sites in a structure layer by layer.

        Args:
            tol (float): tolerance factor when determine whether two atoms are
                are at the same plane. Angstrom
            axis (int): The direction of top and bottom layers. 0: x, 1: y, 2: z

        Returns:
            Lists with the sites in the same plane as one list.
        """
        new_atoms = sorted(self, key=lambda x: x.frac_coords[axis])
        layers = []
        for k, g in groupby(new_atoms, key=lambda x: x.frac_coords[axis]):
            layers.append(list(g))
        new_layers = []
        k = -1
        for i in range(len(layers)):
            if i > k:
                tmp = layers[i]
                for j in range(i + 1, len(layers)):
                    if self.lattice.abc[axis] * abs(
                                    layers[j][0].frac_coords[axis] -
                                    layers[i][0].frac_coords[axis]) < tol:
                        tmp.extend(layers[j])
                        k = j
                    else:
                        break
                new_layers.append(sorted(tmp))
        # check if the 1st layer and last layer are actually the same layer
        # use the fractional as cartesian doesn't work for unorthonormal
        if self.lattice.abc[axis] * abs(
                                new_layers[0][0].frac_coords[axis] + 1 -
                                new_layers[-1][0].frac_coords[axis]) < tol:
            tmp = new_layers[0] + new_layers[-1]
            new_layers = new_layers[1:-1]
            new_layers.append(sorted(tmp))
        return new_layers

    def set_orthogonal_grain(self):
        a, b, c = self.lattice.abc
        self.lattice = Lattice.orthorhombic(a, b, c)

    def build_grains(self, csl, gb_direction, uc_a=1, uc_b=1):
        """
        Build structures for grain A and B from CSL matrix, number of unit cell
        of grain A and number of unit cell of grain B. Each grain is essentially
        a supercell for the initial structure.

        Args:
            csl (3x3 matrix): CSL matrix (scaling matrix)
            gb_direction (int): The direction of GB. 0: x, 1: y, 2: z
            uc_a (int): Number of unit cell of grain A. Default to 1.
            uc_b (int): Number of unit cell of grain B. Default to 1.

        Returns:
            Grain objects for grain A and B
        """
        csl_t = csl.transpose()
        # rotate along a longer axis between a and b
        grain_a = self.copy()
        grain_a.make_supercell(csl_t)
        # grain_a.to(filename='POSCAR')
        # exit()

        if not all([i == 90 for i in grain_a.lattice.angles]):
            warnings.warn("The lattice system of the grain is not orthogonal. "
                          "aimsgb will find a supercell of the grain structure "
                          "that is orthogonalized. This may take a while. ")
            cst = CubicSupercellTransformation(force_90_degrees=True,
                                               min_length=min(grain_a.lattice.abc))
            _s = grain_a.copy()
            _s = cst.apply_transformation(_s)
            _matrix = [reduce_vector(i) for i in cst.transformation_matrix]
            _s = grain_a.copy()
            _s.make_supercell(_matrix)
            sm = StructureMatcher(attempt_supercell=True, primitive_cell=False)
            _matrix = sm.get_supercell_matrix(_s, self)
            matrix = [reduce_vector(i) for i in _matrix]
            grain_a = self.copy()
            grain_a.make_supercell(matrix)
            # grain_a.make_supercell(get_sc_fromstruct(grain_a, min_length=min(grain_a.lattice.abc),
            #                                          force_diagonal=True))
            # grain_a.make_supercell(get_sc_fromstruct(grain_a).transpose())
            # grain_a.set_orthogonal_grain()
            # grain_b = grain_b.set_orthogonal_grain()

        # grain_a.to(filename='POSCAR')
        # exit()
        temp_a = grain_a.copy()
        scale_vector = [1, 1]
        scale_vector.insert(gb_direction, uc_b)
        temp_a.make_supercell(scale_vector)
        grain_b = self.get_b_from_a(temp_a, csl)
        # make sure that all fractional coordinates that equal to 1 will become 0
        grain_b.make_supercell([1, 1, 1])

        scale_vector = [1, 1]
        scale_vector.insert(gb_direction, uc_a)
        grain_a.make_supercell(scale_vector)

        # grain_b.to(filename='POSCAR')
        # exit(0)
        return grain_a, grain_b



#!/usr/bin/env python
from __future__ import division

import sys
import argparse
import numpy as np
from mp_api.client import MPRester, MPRestError

from aimsgb import Grain, GBInformation, GrainBoundary

__author__ = "Jianli Cheng and Kesong YANG"
__copyright__ = "Copyright (C) 2018 The Regents of the University of California"
__maintainer__ = "Jianli Cheng"
__email__ = "jic198@ucsd.edu"
__status__ = "Production"
__date__ = "January 26, 2018"
aimsgb_descript = """**************************************************************************************************
       aimsgb - Ab-initio Interface Materials Simulation for Grain Boundaries (aimsgb.org)      
                        VERSION 0.1  2017 - 2018                                    
Aimsgb is one open-source python library to generate periodic grain boundary structures.    
A reference for the usage of aimsgb software is:
Jianli Cheng, Jian Luo, and Kesong Yang, Aimsgb: An Algorithm and Open-Source Python Library  
to Generate Periodic Grain Boundary Structures, Comput. Mater. Sci. 155, 92-103, (2018). 
**************************************************************************************************

Aimsgb Command Line Tools 
"""


def gb_list(args):
    axis = list(map(int, args.axis))
    sigma = args.sigma
    print(GBInformation(axis, sigma).__str__())


def gb(args):
    axis = list(map(int, args.axis))
    plane = args.plane
    initial_struct = None
    if args.filename:
        initial_struct = Grain.from_file(args.filename)
    elif args.mpid:
        mpr = MPRester()
        try:
            s = mpr.get_structure_by_material_id(args.mpid)
            initial_struct = Grain.from_dict(s.as_dict())
        except MPRestError:
            raise MPRestError("Please run 'export MP_API_KEY=YOUR_MAPI_KEY' to setup your api_key first.")
    if initial_struct is None:
        raise ValueError("Please provide either filename or mpid.")
    
    gb = GrainBoundary(axis, args.sigma, plane, initial_struct, args.uc_a, args.uc_b)
    to_primitive = False if args.conventional else True
    gb.build_gb(args.vacuum, args.add_if_dist, to_primitive, args.delete_layer,
                args.tol).to(filename=args.out, fmt=args.fmt)
    print("CSL Matrix (det=%.1f):\n%s" % (np.linalg.det(gb.csl), gb.csl))
    print("%s of sigma%s[%s]/(%s) is created"
          % (args.out, gb.sigma, args.axis, gb.plane_str))


def main():
    parser = argparse.ArgumentParser(description=aimsgb_descript,
                                     formatter_class=argparse.RawTextHelpFormatter)
    subparsers = parser.add_subparsers(help="command", dest="command")

    parser_gb_list = subparsers.add_parser(
        "list", help="Show the values of sigma, theta, GB plane and CSL matrix "
        "from a given rotation axis\nEXAMPLE: aimsgb list 001",
        formatter_class=argparse.RawTextHelpFormatter)
    parser_gb_list.add_argument("axis", metavar="rotation axis", type=str,
                                help="The rotation axis of GB\n"
                                "EXAMPLE: aimsgb list 001")
    parser_gb_list.add_argument("sigma", default=30, type=int, nargs="?",
                                help="Set the sigma limit for on screen output "
                                     "(default: %(default)s)\n"
                                     "EXAMPLE: aimsgb list 001 100")
    parser_gb_list.set_defaults(func=gb_list)

    parser_gb = subparsers.add_parser(
        "gb", help="Build grain boundary based on rotation axis, sigma, GB plane, "
                   "and input structure file.\nThe user can also specify many "
                   "optional arguments, such grain size and interface terminations."
                   "\nEXAMPLE1: aimsgb gb 001 5 1 2 0 -f POSCAR"
                   "\nEXAMPLE2: aimsgb gb 001 5 1 2 0 -mp mp-13 -ua 2 -ub 3"
                   "\nEXAMPLE3: aimsgb gb 001 5 1 2 0 -f POSCAR -ua 3 -ub 2 -dl 0b1t1b0t",
        formatter_class=argparse.RawTextHelpFormatter)
    parser_gb.add_argument("axis", metavar="rotation axis", type=str,
                           help="The rotation axis of GB, EXAMPLE: 110")
    parser_gb.add_argument("sigma", type=int,
                           help="The sigma value for grain boundary, EXAMPLE: 3")
    parser_gb.add_argument("plane", type=int, nargs=3,
                           help="The GB plane for grain boundary, EXAMPLE: 1 1 0")
    parser_gb.add_argument("-f", "--filename", type=str,
                           help="The initial structure file for grain boundary.")
    parser_gb.add_argument("-mp", "--mpid", type=str,
                           help="The mpid to get initial structure from Materials Project.")
    parser_gb.add_argument("-o", "--out", type=str, default="POSCAR", 
                           help="The output filename. (default: %(default)s)")
    parser_gb.add_argument("-ua", "--uc_a", default=1, type=int,
                           help="The size (uc) for grain A. (default: %(default)s)"
                                "\nEXAMPLE: aimsgb gb 001 5 1 2 0 -f POSCAR -ua 2")
    parser_gb.add_argument("-ub", "--uc_b", default=1, type=int,
                           help="The size (uc) for grain B. (default: %(default)s)"
                                "\nEXAMPLE: aimsgb gb 001 5 1 2 0 -f POSCAR -ub 2")
    parser_gb.add_argument("-dl", "--delete_layer", default="0b0t0b0t", type=str,
                           help="Delete bottom or top layers for each grain. "
                                "(default: %(default)s)"
                                "\nEXAMPLE: aimsgb gb 001 5 1 2 0 -f POSCAR -dl 0b1t1b0t")
    parser_gb.add_argument("-v", "--vacuum", default=0.0, type=float,
                           help="Set vacuum thickness for grain boundary. "
                                "(default: %(default)s angstrom)"
                                "\nEXAMPLE: aimsgb gb 001 5 1 2 0 -f POSCAR -v 20")
    parser_gb.add_argument("-t", "--tol", default=0.25, type=float,
                           help="Tolerance factor to determine if two "
                                "atoms are at the same plane. "
                                "(default: %(default)s angstrom)"
                                "\nEXAMPLE: aimsgb gb 001 5 1 2 0 -f POSCAR -dl 0b1t1b0t -t 0.5")
    parser_gb.add_argument("-ad", "--add_if_dist", default=0.0, type=float,
                           help="Add extra distance between two grains. "
                                "(default: %(default)s angstrom)"
                                "\nEXAMPLE: aimsgb gb 001 5 1 2 0 -f POSCAR -ad 0.5")
    parser_gb.add_argument("-c", "--conventional", action="store_true",
                           help="Get conventional GB, not primitive"
                                "\nEXAMPLE: aimsgb gb 001 5 1 2 0 -f POSCAR -c")
    parser_gb.add_argument("-fmt", "--fmt", default="poscar", const="poscar",
                           nargs="?", choices=["poscar", "cif", "cssr", "json"],
                           help="Choose the output format. (default: %(default)s)"
                                "\nEXAMPLE: aimsgb gb 001 5 1 2 0 -f POSCAR -fmt cif")
    parser_gb.set_defaults(func=gb)

    args = parser.parse_args()

    try:
        getattr(args, "func")
    except AttributeError:
        parser.print_help()
        sys.exit(0)
    args.func(args)


if __name__ == "__main__":
    main()



from __future__ import division

import numpy as np
from numpy.linalg import norm
from functools import wraps, reduce
try:
    from math import gcd as pygcd
except ImportError:
    from fractions import gcd as pygcd

__author__ = "Jianli Cheng, Kesong Yang"
__copyright__ = "Copyright 2018, Yanggroup"
__maintainer__ = "Jianli Cheng"
__email__ = "jic198@ucsd.edu"
__status__ = "Production"
__date__ = "January 26, 2018"


def reduce_vector(vector):
    """
    Reduce a vector
    """
    d = abs(reduce(gcd, vector))
    vector = tuple([int(i / d) for i in vector])

    return vector


def co_prime(a, b):
    """
    Check if two integers are co-prime
    """
    return gcd(a, b) in (0, 1)


def plus_minus_gen(start, end):
    """
    Generate a list of plus and minus alternating integers
    """
    for i in range(start, end):
        yield i
        yield -i


def is_integer(a, tol=1e-5):
    """
    Check whether the number is integer
    """
    return norm(abs(a - np.round(a))) < tol


def get_smallest_multiplier(a, max_n=10000):
    """
    Get the smallest multiplier to make the list with all integers
    Args:
        a (list): A list of numbers
        max_n (int): The up limit to search multiplier

    Returns:
        The smallest integer multiplier
    """
    a = np.array(a)
    for i in range(1, max_n):
        if is_integer(i * a):
            return i
    raise ValueError("Cannot find an integer matrix with multiplier "
                     "searched already up to %s" % max_n)


def reduce_integer(integer):
    """
    Get the odd number for an integer
    """
    while gcd(integer, 2) != 1:
        integer //= 2
    return integer


def gcd(*numbers):
    """
    Get a greatest common divisor for a list of numbers
    """
    n = numbers[0]
    for i in numbers:
        n = pygcd(n, i)
    return n


def transpose_matrix(func):
    """
    Transpose the first argument and the return value.
    Args:
        func: The function that uses transpose_matrix as a decorator.
    """

    @wraps(func)
    def transpose(*args, **kwargs):
        args_list = list(args)
        args_list[0] = args_list[0].transpose()
        matrix = func(*args_list, **kwargs)
        return matrix.transpose()

    return transpose



from setuptools import setup, find_packages


long_desc = """
Introduction
============
aimsgb, an efficient and open-source Python library for generating atomic coordinates in periodic grain boundary models. It is designed to
construct various grain boundary structures from cubic and non-cubic initial
configurations. A convenient command line tool has also been provided to enable
easy and fast construction of tilt and twist boundaries by assigining the degree
of fit (Σ), rotation axis, grain boundary plane and initial crystal structure.
aimsgb is expected to greatly accelerate the theoretical investigation of
grain boundary properties and facilitate the experimental analysis of grain
boundary structures as well.

A reference for the usage of aimsGB software is:
Jianli Cheng, Jian Luo, and Kesong Yang, Aimsgb: An Algorithm and OPen-Source Python Library to Generate Periodic Grain Boundary Structures, Comput. Mater. Sci. 155, 92-103, (2018). 
DOI:10.1016/j.commatsci.2018.08.029  

Install aimsgb
==============
1. Clone the latest version from github::

    git clone git@github.com:ksyang2013/aimsgb.git

2. Navigate to aimsgb folder::

    cd aimsgb

3. Type in the root of the repo::

    pip install .

4. or to install the package in development mode::

    pip install -e .


How to cite aimsgb
==================

If you use aimsgb in your research, please consider citing the following work:

    Jianli Cheng, Jian Luo, Kesong Yang. *Aimsgb: An algorithm and open-source python
    library to generate periodic grain boundary structures.* Computational Materials
    Science, 2018, 155, 92-103. `doi:10.1016/j.commatsci.2018.08.029
    <https://doi.org/10.1016/j.commatsci.2018.08.029>`_


Copyright
=========
Copyright (C) 2018 The Regents of the University of California

All Rights Reserved. Permission to copy, modify, and distribute this software and its documentation for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies. Permission to make commercial use of this software may be obtained by contacting:

Office of Innovation and Commercialization
9500 Gilman Drive, Mail Code 0910
University of California
La Jolla, CA 92093-0910
(858) 534-5815
innovation@ucsd.edu

This software program and documentation are copyrighted by The Regents of the University of California. The software program and documentation are supplied “as is”, without any accompanying services from The Regents. The Regents does not warrant that the operation of the program will be uninterrupted or error-free. The end-user understands that the program was developed for research purposes and is advised not to rely exclusively on the program for any reason.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN “AS IS” BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.


Authors
=======
Dr. Jianli Cheng (jic198@ucsd.edu)
New Email: jianlicheng@lbl.gov

Prof. Kesong Yang  (kesong@ucsd.edu)

About the aimsgb Development Team
=================================
http://materials.ucsd.edu/
"""

setup(
    name="aimsgb",
    packages=find_packages(),
    version="0.1.2",
    setup_requires=["setuptools>=18.0"],
    install_requires=["pymatgen", "mp_api", "numpy"],
    include_package_data=True,
    author="Jianli Cheng and Kesong YANG",
    maintainer="Jianli Cheng, Sicong JIANG, and Kesong YANG",
    maintainer_email="chengjianli90@gmail.com, sij014@ucsd.edu, kesong@ucsd.edu",
    url="http://aimsgb.org",
    description="aimsgb is a python library for generatng the atomic "
                "coordinates of periodic grain boundaries."
                "Copyright © 2018 The Regents of the University of California."
                "All Rights Reserved. See more in Copyright.",
    long_description=long_desc,
    keywords=["material science", "grain boundary", "molecular simulation"],
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    entry_points={
        'console_scripts': [
            'aimsgb = aimsgb.agb:main',
        ]
    }
)
